@@ -1224,17 +1224,11 @@ const char * Hunspell::get_version()
   return pAMgr->get_version();
 }
 
-/*
-
 struct cs_info * Hunspell::get_csconv()
 {
   return csconv;
 }
 
-int utf16_isalpha(unsigned short c) {
-  return unicodeisalpha(c);
-}
-*/
 #ifdef HUNSPELL_EXPERIMENTAL
 // XXX need UTF-8 support
 char * Hunspell::morph(const char * word)
