@@ -89,9 +89,9 @@ LaTeXParser::LaTeXParser(const char * wordchars)
 	init(wordchars);
 }
 
-LaTeXParser::LaTeXParser(struct unicode_info2 * uconv, unsigned short * wordchars, int len)
+LaTeXParser::LaTeXParser(unsigned short * wordchars, int len)
 {
-	init(uconv, wordchars, len);
+	init(wordchars, len);
 }
 
 LaTeXParser::~LaTeXParser() 
