@@ -20,9 +20,9 @@ ManParser::ManParser(const char * wordchars)
 	// wordcharacters[('/' + 256) % 256] = 1; // for paths
 }
 
-ManParser::ManParser(struct unicode_info2 * uconv, unsigned short * wordchars, int len)
+ManParser::ManParser(unsigned short * wordchars, int len)
 {
-	init(uconv, wordchars, len);
+	init(wordchars, len);
 }
 
 ManParser::~ManParser() 
